@@ -92,8 +92,8 @@ typedef int IFASLCODE;      /* fasl type codes */
 
 #define find_room(tc, s, g, t, n, x) find_gc_room_T(THREAD_GC(tc), s, g, t, n, ALREADY_PTR, x)
 #define find_gc_room(tgc, s, g, t, n, x) find_gc_room_T(tgc, s, g, t, n, ALREADY_PTR, x)
-#define find_room_voidp(tc, s, g, n, x) find_gc_room_T(THREAD_GC(tc), s, g, typemod, n, TO_VOIDP, x)
-#define find_gc_room_voidp(tgc, s, g, n, x) find_gc_room_T(tgc, s, g, typemod, n, TO_VOIDP, x)
+#define find_room_voidp(tc, s, g, n, x) find_gc_room_T(THREAD_GC(tc), s, g, type_untyped, n, TO_VOIDP, x)
+#define find_gc_room_voidp(tgc, s, g, n, x) find_gc_room_T(tgc, s, g, type_untyped, n, TO_VOIDP, x)
 
 /* new-space inline allocation --- no mutex required */
 /* Like `find_room`, but always `space_new` and generation 0,
@@ -111,7 +111,7 @@ typedef int IFASLCODE;      /* fasl type codes */
  } while(0)
 
 #define newspace_find_room(tc, t, n, x) newspace_find_room_T(tc, t, n, ALREADY_PTR, x)
-#define newspace_find_room_voidp(tc, n, x) newspace_find_room_T(tc, typemod, n, TO_VOIDP, x)
+#define newspace_find_room_voidp(tc, n, x) newspace_find_room_T(tc, type_untyped, n, TO_VOIDP, x)
 
 #ifndef NO_PRESERVE_FLONUM_EQ
 # define PRESERVE_FLONUM_EQ
@@ -158,6 +158,9 @@ typedef struct _seginfo {
   struct _seginfo *next;                    /* pointer to the next seginfo (used in occupied_segments and unused_segs) */
   struct _seginfo *sweep_next;              /* next in list of segments allocated during GC => need to sweep */
   ptr sweep_start;                          /* address within segment to start sweep */
+#if defined(WRITE_XOR_EXECUTE_CODE)
+  iptr sweep_bytes;                         /* total number of bytes starting at sweep_start */
+#endif
   struct _seginfo **dirty_prev;             /* pointer to the next pointer on the previous seginfo in the DirtySegments list */
   struct _seginfo *dirty_next;              /* pointer to the next seginfo on the DirtySegments list */
   ptr trigger_ephemerons;                   /* ephemerons to re-check if object in segment is copied out */
@@ -170,6 +173,9 @@ typedef struct _seginfo {
   octet *counting_mask;                     /* bitmap of counting roots during a GC */
   octet *measured_mask;                     /* bitmap of objects that have been measured */
 #ifdef PORTABLE_BYTECODE
+# ifndef PTHREADS
+  void *encorage_alignment;                 /* hack for 32-bit systems that align 64-bit values on 4 bytes */
+# endif
   union { ptr force_alignment;    
 #endif
   octet dirty_bytes[cards_per_segment];     /* one dirty byte per card */
@@ -183,7 +189,7 @@ typedef struct _chunkinfo {
   iptr base;                                /* first segment */
   iptr bytes;                               /* size in bytes */
   iptr segs;                                /* size in segments */
-  iptr nused_segs;                          /* number of segments currently in used use */ 
+  iptr nused_segs;                          /* number of segments currently in use */
   struct _chunkinfo **prev;                 /* pointer to previous chunk's next */
   struct _chunkinfo *next;                  /* next chunk */
   struct _seginfo *unused_segs;             /* list of unused segments */
@@ -268,6 +274,7 @@ typedef struct _bucket_pointer_list {
 #define size_closure(n) ptr_align(header_size_closure + (n)*ptr_bytes)
 #define size_string(n) ptr_align(header_size_string + (n)*string_char_bytes)
 #define size_fxvector(n) ptr_align(header_size_fxvector + (n)*ptr_bytes)
+#define size_flvector(n) ptr_align(header_size_flvector + (n)*sizeof(double))
 #define size_bytevector(n) ptr_align(header_size_bytevector + (n))
 #define size_bignum(n) ptr_align(header_size_bignum + (n)*bigit_bytes)
 #define size_code(n) ptr_align(header_size_code + (n))
@@ -275,6 +282,8 @@ typedef struct _bucket_pointer_list {
 #define size_record_inst(n) ptr_align(n)
 #define unaligned_size_record_inst(n) (n)
 
+#define rtd_parent(x) INITVECTIT(RECORDDESCANCESTRY(x), Svector_length(RECORDDESCANCESTRY(x)) - ancestry_parent_offset)
+
 /* type tagging macros */
 
 #define TYPE(x,type) ((ptr)((iptr)(x) - typemod + (type)))
@@ -321,17 +330,15 @@ typedef struct _dirtycardinfo {
 #define ENTRYFRAMESIZE(x) (ISENTRYCOMPACT(x)                            \
                            ? ((COMPACTENTRYFIELD(x, compact_frame_words_offset) & compact_frame_words_mask) << log2_ptr_bytes) \
                            : RPHEADERFRAMESIZE((uptr)(x) - size_rp_header))
-#define ENTRYOFFSET(x) (ISENTRYCOMPACT(x)                               \
-                        ? RPCOMPACTHEADERTOPLINK((uptr)(x) - size_rp_compact_header) \
-                        : RPHEADERTOPLINK((uptr)(x) - size_rp_header))
-#define ENTRYOFFSETADDR(x) (ISENTRYCOMPACT(x)                               \
-                            ? &RPCOMPACTHEADERTOPLINK((uptr)(x) - size_rp_compact_header) \
-                            : &RPHEADERTOPLINK((uptr)(x) - size_rp_header))
 #define ENTRYLIVEMASK(x) (ISENTRYCOMPACT(x)                             \
                           ? FIX(COMPACTENTRYFIELD(x, compact_frame_mask_offset)) \
                           : RPHEADERLIVEMASK((uptr)(x) - size_rp_header))
 #define ENTRYNONCOMPACTLIVEMASKADDR(x) (&RPHEADERLIVEMASK((uptr)(x) - size_rp_header))
 
+/* `top-link` must be a fixed distance from end or header, whether compact or not: */
+#define ENTRYOFFSET(x) (RPCOMPACTHEADERTOPLINK((uptr)(x) - size_rp_compact_header))
+#define ENTRYOFFSETADDR(x) (&ENTRYOFFSET(x))
+
 #define PORTFD(x) ((iptr)PORTHANDLER(x))
 #define PORTGZFILE(x) ((gzFile)(PORTHANDLER(x)))
 
@@ -353,7 +360,7 @@ typedef struct {
   s_thread_mutex_t pmutex;
 } scheme_mutex_t;
 
-#define get_thread_context() (ptr)s_thread_getspecific(S_tc_key)
+#define get_thread_context() TO_PTR(s_thread_getspecific(S_tc_key))
 /* deactivate thread prepares the thread for a possible collection.
    if it's the last active thread, it signals one of the threads
    waiting on the collect condition, if any, so that a collection
@@ -391,11 +398,9 @@ typedef struct {
     tc_mutex_release()\
   }\
 }
-/* S_tc_mutex_depth records the number of nested mutex acquires in
-   C code on tc_mutex.  it is used by do_error to release tc_mutex
-   the appropriate number of times.
-*/
+
 #define tc_mutex_acquire() do {                 \
+    assert_no_alloc_mutex();                    \
     S_mutex_acquire(&S_tc_mutex);               \
     S_tc_mutex_depth += 1;                      \
   } while (0);
@@ -403,9 +408,31 @@ typedef struct {
     S_tc_mutex_depth -= 1;                      \
     S_mutex_release(&S_tc_mutex);               \
   } while (0);
-#define gc_tc_mutex_acquire() S_mutex_acquire(&S_gc_tc_mutex)
-#define gc_tc_mutex_release() S_mutex_release(&S_gc_tc_mutex)
 
+/* Allocation mutex is ordered after tc mutex */
+#define alloc_mutex_acquire() do {              \
+    S_mutex_acquire(&S_alloc_mutex);            \
+    S_alloc_mutex_depth += 1;                   \
+  } while (0);
+#define alloc_mutex_release() do {              \
+    S_alloc_mutex_depth -= 1;                   \
+    S_mutex_release(&S_alloc_mutex);            \
+  } while (0);
+
+/* To enable checking lock order: */
+#if 0
+# define assert_no_alloc_mutex() do {                                   \
+    if (S_mutex_is_owner(&S_alloc_mutex))                               \
+      S_error_abort("cannot take tc mutex after allocation mutex");     \
+  } while (0)
+#else
+# define assert_no_alloc_mutex() do { } while (0)
+#endif
+
+#define IS_TC_MUTEX_OWNER() S_mutex_is_owner(&S_tc_mutex)
+#define IS_ALLOC_MUTEX_OWNER() S_mutex_is_owner(&S_alloc_mutex)
+
+/* Enable in "version.h": */
 #ifdef IMPLICIT_ATOMIC_AS_EXPLICIT
 # define AS_IMPLICIT_ATOMIC(T, X) ({       \
       T RESLT;                             \
@@ -434,8 +461,10 @@ typedef struct {
 #define reactivate_thread(tc) {}
 #define tc_mutex_acquire() do {} while (0)
 #define tc_mutex_release() do {} while (0)
-#define gc_tc_mutex_acquire() do {} while (0)
-#define gc_tc_mutex_release() do {} while (0)
+#define alloc_mutex_acquire() do {} while (0)
+#define alloc_mutex_release() do {} while (0)
+#define IS_TC_MUTEX_OWNER() 1
+#define IS_ALLOC_MUTEX_OWNER() 1
 #define S_cas_load_acquire_voidp(a, old, new) (*(a) = new, 1)
 #define S_cas_store_release_voidp(a, old, new) (*(a) = new, 1)
 #define S_cas_load_acquire_ptr(a, old, new) (*(a) = new, 1)
@@ -446,17 +475,12 @@ typedef struct {
 #define AS_IMPLICIT_ATOMIC(T, X) X
 #endif
 
-typedef struct remote_range {
-  ISPC s;
-  IGEN g;
-  ptr start, end;
-  struct thread_gc *tgc;
-  struct remote_range *next;
-} remote_range;
-
 typedef struct thread_gc {
   ptr tc;
-  ptr thread; /* set only when collecting */
+
+  int during_alloc;
+  IBOOL queued_fire;
+  IBOOL preserve_ownership;
 
   struct thread_gc *next;
 
@@ -476,12 +500,16 @@ typedef struct thread_gc {
   int sweep_change;
   
   int sweeper; /* parallel GC: sweeper thread identity */
-  
-  struct thread_gc *remote_range_tgc;
-  ptr remote_range_start;
-  ptr remote_range_end;
-  remote_range *ranges_to_send; /* modified only by owning sweeper */
-  remote_range *ranges_received; /* modified with sweeper mutex held */
+
+  /* modified only by owning sweeper; contains ptr and thread_gc* */
+  ptr send_remote_sweep_stack;
+  ptr send_remote_sweep_stack_start;
+  ptr send_remote_sweep_stack_limit;
+
+  /* modified with sweeper mutex held; contains just ptr */
+  ptr receive_remote_sweep_stack;
+  ptr receive_remote_sweep_stack_start;
+  ptr receive_remote_sweep_stack_limit;
 
   seginfo *dirty_segments[DIRTY_SEGMENT_LISTS];
 
@@ -492,7 +520,7 @@ typedef struct thread_gc {
 
 #define main_sweeper_index maximum_parallel_collect_threads
 
-#ifdef __MINGW32__
+#if defined(__MINGW32__) && !defined(HAND_CODED_SETJMP_SIZE)
 /* With MinGW on 64-bit Windows, setjmp/longjmp is not reliable. Using
    __builtin_setjmp/__builtin_longjmp is reliable, but
    __builtin_longjmp requires 1 as its second argument. So, allocate
@@ -502,6 +530,11 @@ typedef struct thread_gc {
 # define FREEJMPBUF(jb) free(jb)
 # define SETJMP(jb) (JMPBUF_RET(jb) = 0, __builtin_setjmp(jb), JMPBUF_RET(jb))
 # define LONGJMP(jb,n) (JMPBUF_RET(jb) = n, __builtin_longjmp(jb, 1))
+#elif defined(HAND_CODED_SETJMP_SIZE)
+# define CREATEJMPBUF() malloc(sizeof(ptr) * HAND_CODED_SETJMP_SIZE)
+# define FREEJMPBUF(jb) free(jb)
+# define SETJMP(jb) S_setjmp(jb)
+# define LONGJMP(jb,n) S_longjmp(jb, n)
 #else
 /* assuming malloc will give us required alignment */
 # define CREATEJMPBUF() malloc(sizeof(jmp_buf))
@@ -513,7 +546,7 @@ typedef struct thread_gc {
 #define DOUNDERFLOW\
  &CODEIT(CLOSCODE(S_lookup_library_entry(library_dounderflow, 1)),size_rp_header)
 
-#define HEAP_VERSION_LENGTH 16
+#define HEAP_VERSION_LENGTH 24
 #define HEAP_MACHID_LENGTH 16
 #define HEAP_STAMP_LENGTH 16
 
@@ -526,9 +559,20 @@ typedef struct thread_gc {
 #define SETPTRFIELD(x,disp,y) DIRTYSET(((ptr *)TO_VOIDP((uptr)(x)+disp)),(y))
 
 #define INCRGEN(g) (g = g == S_G.max_nonstatic_generation ? static_generation : g+1)
-#define IMMEDIATE(x) (Sfixnump(x) || Simmediatep(x))
+#define FIXMEDIATE(x) (Sfixnump(x) || Simmediatep(x))
+
+#define Sbytevector_reference_length(p) (Sbytevector_length(p) >> log2_ptr_bytes)
+#define INITBVREFIT(p, i) (*(ptr *)(&BVIT(p, (i) << log2_ptr_bytes)))
 
 /* For `memcpy_aligned, that the first two arguments are word-aligned
    and it would be ok to round up the length to a word size. But
    probably the compiler does a fine job with plain old `mempcy`. */
 #define memcpy_aligned memcpy
+
+#define USE_TRAP_FUEL(tc, n) do {                         \
+    uptr _amt_ = (uptr)(n);                               \
+    if ((uptr)TRAP(tc) > _amt_)                           \
+      TRAP(tc) = (ptr)((uptr)TRAP(tc) - _amt_);           \
+     else                                                 \
+       TRAP(tc) = (ptr)1;                                 \
+  } while (0)
